from ...imports import *

__all__ = ["inject_noise"]


def inject_noise(self, signal_to_noise=100):
    """
    Inject uncorrelated Gaussian random noise into the flux.

    Parameters
    ----------
    signal_to_noise : float
        The signal-to-noise per wavelength per time.
        For example, S/N=100 would mean that the
        uncertainty on the flux for each each
        wavelength-time data point will be 1%.

    Returns
    -------
    rainbow : Rainbow
        A new Rainbow object with the systematics injected.
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("inject_noise", locals())

    # create a copy of the existing Rainbow
    new = self._create_copy()

    # store S/N as metadata
    new.metadata["signal_to_noise"] = signal_to_noise

    # get the underlying model (or create one if needed)
    if "model" in new.fluxlike:
        model = new.fluxlike["model"]
    else:
        # kludge, do we really want to allow this?
        model = self.flux * 1
        new.fluxlike["model"] = model

    # calculate the uncertainty with a fixed S/N
    uncertainty = model / signal_to_noise
    new.fluxlike["uncertainty"] = uncertainty

    # inject a realization of the noise
    new.fluxlike["flux"] = np.random.normal(model, uncertainty)

    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the new object
    return new



from .rainbow import *
from .visualizations import _add_panel_labels


class RainbowWithModel(Rainbow):
    """
    `RainbowWithModel` objects have a fluxlike `model`
    attached to them, meaning that they can

    This class definition inherits from `Rainbow`.
    """

    @property
    def residuals(self):
        """
        Calculate the residuals on the fly,
        to make sure they're always up to date.
        """
        return self.flux - self.model

    def _validate_core_dictionaries(self):
        super()._validate_core_dictionaries()
        try:
            model = self.get("model")
            assert np.shape(model) == np.shape(self.flux)
        except (AttributeError, AssertionError):
            message = """
            No fluxlike 'model' was found attached to this
            `RainbowWithModel` object. The poor thing,
            its name is a lie! Please connect a model.
            The simplest way to do so might look like...
            `rainbow.model = np.ones(rainbow.shape)`
            ...or similarly with a more interesting array.
            """
            warnings.warn(message)

    def imshow_data_with_models(
        self,
        models=["model"],
        vlimits_data=[0.98, 1.02],
        vspan_residuals=0.02,
        figsize=(8, 3),
        label="inside",
        labelkw={},
        **kw,
    ):
        """
        Produce a multi-panel imshow figure comparing
        the data to model components. The left panel
        will be the data ('flux'), the right panel will
        be the residuals ('flux' - 'model'), and the middle
        panel(s) will be the model(s) connecting them.

        Parameters
        ----------
        models : list
            The list of model quantities to display,
            each in its own panel. A popular option might be:
            `models = ['systematics_model', 'planet_model']
            Whatever you include here does not affect the
            values of the Residuals panel, which are always
            exactly 'flux' - 'model'.
        vlimits_data : list
            The value limits for the data and model colormap,
            specified as `vlimits_data = [vmin, vmax]`.
            If `None` or `[None, None]`, the limits will be
            replaced with the [1,99] percentiles of the flux.
        vspan_residuals : float
            The distance away from zero where the colormap
            for the residuals will cut off. If `None`, the
            span will be set to 3 standard deviations.
        figsize : tuple
            The figure size for the multipanel plot.
            It should probably be wider than it is tall.
        label : bool
            Should we add (a), (b), (c) labels to the panels,
            and where? False or None give no labels, 'inside'
            sets them in the upper left inside corner, 'outside'
            sets them above the upper left corner.
        labelkw : dict
            Keywords to pass along to the `plt.text` command
            that makes the labels.
        kw : dict
            All additional keywords will be passed along
            to `Rainbow.imshow` for all imshow panels.
        """
        # make sure color limits are set for data and model
        percentiles = [1, 99]
        if vlimits_data is None:
            vlimits_data = np.percentile(self.flux, percentiles)
        elif vlimits_data[0] is None:
            vlimits_data[0] = np.percentile(self.flux, percentiles[0])
        elif vlimits_data[1] is None:
            vlimits_data[1] = np.percentile(self.flux, percentiles[1])

        # make sure color limits are set for the residuals
        if vspan_residuals is None:
            vspan_residuals = np.nanstd(self.residuals) * 3

        # set up the panels
        row_data, row_colorbar = 1, 0
        N = len(models) + 2
        fi, ax = plt.subplots(
            2,
            N,
            figsize=figsize,
            dpi=600,
            gridspec_kw=dict(height_ratios=[1, 20]),
            sharex="row",
            sharey="row",
            constrained_layout=True,
        )

        # a helper function to make better titles
        def create_title(s):
            if s == "flux":
                return "Data"
            else:
                return s.replace("_", " ").title()

        # plot the first three panels
        for i, k in enumerate(["flux"] + models):
            self.imshow(
                ax=ax[row_data, i],
                quantity=k,
                vmin=vlimits_data[0],
                vmax=vlimits_data[1],
                colorbar=False,
                **kw,
            )
            ax[row_data, i].set_title(create_title(k))

        # set up the colorbar to span the data and models
        gs = ax[row_colorbar, 0].get_gridspec()
        for a in ax[row_colorbar, :-1]:
            a.remove()
        ax_colorbar = fi.add_subplot(gs[row_colorbar, :-1])
        plt.colorbar(cax=ax_colorbar, orientation="horizontal")

        # plot the residuals
        k = "residuals"
        self.imshow(
            ax[row_data, -1],
            k,
            vmin=-vspan_residuals,
            vmax=vspan_residuals,
            colorbar=False,
            **kw,
        )
        ax[row_data, -1].set_title("Residuals")

        # set up the colorbar for the residuals
        ax_colorbar_residuals = plt.subplot(gs[row_colorbar, -1])
        cbar = plt.colorbar(cax=ax_colorbar_residuals, orientation="horizontal")
        for a in [ax_colorbar, ax_colorbar_residuals]:
            a.tick_params(labelsize=8)
        x = vspan_residuals / 2
        cbar.set_ticks([-x, 0, x], labels=[f"{-x:.2}", "0", f"{x:.2}"])

        # clear all the axis labels and replace with shared
        for a in ax[row_data, :]:
            xlabel, ylabel = a.get_xlabel(), a.get_ylabel()
            a.set_ylabel("")
            a.set_xlabel("")
        fi.supxlabel(xlabel)
        fi.supylabel(ylabel)

        if label:
            _add_panel_labels(ax[row_data, :], preset=label, **labelkw)



from ..rainbows import *
from .setup_tests import *


def test_attach_model():
    simulated = SimulatedRainbow().inject_noise().inject_transit().inject_systematics()
    original = simulated._create_copy()

    # pull out the model
    inputs = {}
    for k in ["model", "planet_model", "systematics_model"]:
        inputs[k] = simulated.fluxlike[k]
        del simulated.fluxlike[k]

    # reattach the model
    withmodel = simulated.attach_model(**inputs)
    assert simulated != withmodel
    assert withmodel == original


def test_imshow_data_with_models():
    s = (
        SimulatedRainbow()
        .inject_transit()
        .inject_systematics()
        .inject_noise()
        .bin(R=50, dt=5 * u.minute)
    )
    s.imshow_data_with_models(cmap="gray")
    plt.savefig(os.path.join(test_directory, "imshow-data-with-model.pdf"))
    s.imshow_data_with_models(models=["systematics_model", "planet_model"], cmap="gray")
    plt.savefig(os.path.join(test_directory, "imshow-data-with-model-components.pdf"))

    s.imshow_data_with_models(models=["systematics_model", "planet_model"], cmap="gray")
    s.imshow_data_with_models(
        models=["systematics_model", "planet_model"], cmap="gray", label=False
    )
    s.imshow_data_with_models(
        models=["systematics_model", "planet_model"],
        cmap="gray",
        labelkw=dict(color="red"),
    )
    s.imshow_data_with_models(
        models=["systematics_model", "planet_model"], cmap="gray", label="outside"
    )



from ....imports import *

__all__ = ["plot_spectral_resolution"]


def plot_spectral_resolution(
    self,
    pixels_per_resolution_element=1,
    ax=None,
    w_unit="micron",
    cmap=None,
    vmin=None,
    vmax=None,
    plotkw={},
    **kw,
):
    """
    Plot the spectral resolution as a function of wavelength.

    Parameters
    ----------
    pixels_per_resolution_element : float
        How many pixels do we consider as a resolution element?
    ax : matplotlib.axes.Axes
        The axes into which to make this plot.
    w_unit : str, astropy.unit.Unit
        The unit for plotting wavelengths.
    cmap : str, matplotlib.colors.Colormap
        The color map to use for expressing wavelength.
    vmin : astropy.units.Quantity
        The minimum value to use for the wavelength colormap.
    vmax : astropy.units.Quantity
        The maximum value to use for the wavelength colormap.
    plotkw : dict
        A dictionary of keywords passed to `plt.plot`
    """
    R = self.get_spectral_resolution(pixels_per_resolution_element)
    w_unit = u.Unit(w_unit)

    with quantity_support():

        # make sure that the wavelength-based colormap is defined
        self._make_sure_cmap_is_defined(cmap=cmap, vmin=vmin, vmax=vmax)

        # make sure ax is set up
        if ax is None:
            ax = plt.subplot()
        plt.sca(ax)

        # define default visuals, but let user override with scatterkw
        # this_scatterkw = dict(c=self.get_wavelength_color(self.wavelength))
        # this_scatterkw.update(**scatterkw)

        plt.plot(self.wavelength.to(w_unit), R, **plotkw)
        plt.xlabel(f'Wavelength ({w_unit.to_string("latex_inline")})')
        plt.ylabel(f"$R=\lambda/d\lambda$ ({pixels_per_resolution_element} pixel)")
        return ax

        ax.set_xlabel(
            f"{self._time_label} ({self.time.unit.to_string('latex_inline')})"
        )



from ...imports import *

__all__ = ["get_measured_scatter"]


def get_measured_scatter(self,quantity='flux', method="standard-deviation"):
    """
    Get measured scatter from a chromatic rainbow.

    Parameters
    ----------
    method : string
        What method to use to obtain measured scatter. Current options are 'MAD', 'standard-deviation'.

    Returns
    -------
    scatter : np.array (wavelike)
        The scatter wavelike array.
    """

    if method == "standard-deviation":
        scatters = np.zeros(len(self.wavelength))
        for i, row in enumerate(self.fluxlike[quantity]):
            scatters[i - 1] = np.std(row)
        return scatters

    if method == "MAD":
        scatters = median_abs_deviation(
            self.fluxlike[quantity].copy(), axis=1, scale="normal"
        )  # Setting the scale 'normal' is like * 1.48
        return scatters

    else:
        warnings.warn(
            f"""
        '{method}' is not an available method.
        Please choose from ['MAD', 'standard-deviation'].
        """
        )



from ...imports import *

__all__ = ["get_spectrum"]


def get_spectrum(self):
    """
    Return a spectrum of the star, averaged over all times.

    Parameters
    ----------

    Returns
    -------
    spectrum : np.array (wavelike)
    """
    return self.get_spectrum_as_rainbow().flux[:, 0]



from ...imports import *

__all__ = ["_add_panel_labels"]


def _add_panel_labels(axes, preset="inside", **kw):
    """
    Add (a), (b), (c) labels to a group of axes.

    Parameters
    ----------
    ax : list or array of matplotlib.axes._subplots.AxesSubplot objects
        The axes into which the labels should be drawn.
    preset : str
        A few presets for where to put the labels relative to
        upper left corner of each panel. Options are ['inside', 'above']
    kw : dict
        All addition keywords will be passed to `plt.text`,
        and they will overwrite defaults.
    """

    textkw = dict(x=0.02, y=0.98, va="top", ha="left")
    if preset == "inside":
        textkw.update(x=0.02, y=0.98, va="top", color="white")
    elif preset == "outside":
        textkw.update(x=0, y=1.02, va="bottom", color="black")
    textkw.update(**kw)

    letters = "abcdefghijklmnopqrstuvwxyz"
    for i, a in enumerate(axes.flatten()):
        textkw["s"] = f"({letters[i]})"
        textkw["transform"] = a.transAxes
        a.text(**textkw)
